class BaseFunction:
	"""

	Examples
	--------
	>>> BaseFunction("SomeFunction", "Text").res
	True or False
	"""

	def __init__(self, command: str, text: str):
		self.command = command.strip()
		self.text = text

		self.check_command()
		self.res = self.execute()

	def check_command(self):
		"""
		Checks command syntax

		Raises
		------
		Exception:
			Any exception from child classes.
		"""
		...

	def execute(self) -> bool:
		"""

		Returns
		-------
		bool
		"""
		...


from .base import BaseFunction
from ..exceptions import ForsetiFunctionSyntaxError


class Length(BaseFunction):
	def __init__(self, command: str, text: str):
		super().__init__(command, text)

	def check_command(self):
		command_prefix = self.command[:3]

		if "|ll" == command_prefix or "|lg" == command_prefix:
			if not self.command[3:].isdigit():
				raise ForsetiFunctionSyntaxError("After |ll or |lg construction must be number")

		else:
			raise ForsetiFunctionSyntaxError("Unsupported command syntax")

	def execute(self) -> bool:
		if "|ll" in self.command:
			return len(self.text) < int(self.command[3:])

		if "|lg" in self.command:
			return len(self.text) > int(self.command[3:])



from typing import List

from ..exceptions import ForsetiSyntaxError


_OPERATORS = ("AND", "OR", "NOT", "И", "ИЛИ")
ERRORS = {
	"operator_position": "Operator '{}' is used without expression '{}'. Position {}",
}


def check_syntax(command: str):
	command = command.replace(" И ", " AND ")
	command = command.replace(" ИЛИ ", " OR ")
	command_parts = command.split()
	parts_number = len(command_parts)

	for i in range(parts_number):
		if command_parts[i] in _OPERATORS:
			word_start_symbol_num = get_word_start_symbol_num(i, command_parts)

			if i == 0:
				raise ForsetiSyntaxError(ERRORS['operator_position'].format(command_parts[i], 'левостороннего', word_start_symbol_num))

			elif i == parts_number - 1:
				raise ForsetiSyntaxError(ERRORS['operator_position'].format(command_parts[i], 'правостороннего', word_start_symbol_num))

			if command_parts[i] == "NOT":
				if command_parts[i - 1] != "AND":
					raise ForsetiSyntaxError("Operator 'NOT' cannot be used without operator 'AND'")

			if command_parts[i] in _OPERATORS:
				if command_parts[i - 1]:
					pass
		else:
			if i < parts_number - 1:
				if command_parts[i + 1].lstrip()[0] == '(' and command_parts[i] not in _OPERATORS:
					raise ForsetiSyntaxError("You can't use staples without operators")

			if i:
				if command_parts[i - 1].rstrip()[-1] == ')' and command_parts[i] not in _OPERATORS:
					raise ForsetiSyntaxError("You can't use staples without operators")


def get_word_start_symbol_num(word_position: int, split_text: List[str]):
	word_start_symbol_num = sum(map(len, split_text[:word_position])) + word_position

	return word_start_symbol_num + 1



from .distance import Distance
from ..exceptions import ForsetiFunctionSyntaxError

import pytest


@pytest.mark.parametrize(
	"condition, text",
	[
		("|wN", ""),
		("|cN", ""),
		("|w5 world", ""),
		("hello |c5", ""),
		("hello |w-1 world", ""),
		("", ""),
	]
)
def test_distance_function_exceptions(condition, text):
	with pytest.raises(ForsetiFunctionSyntaxError):
		Distance(condition, text)


@pytest.mark.parametrize(
	"condition, text, expected_result",
	[
		("Hugin |w5 Munin", "Odin had two ravens Hugin and Munin", True),
		("Hugin |c5 Munin", "Odin had two ravens Hugin and Munin", True),
		("Hugin |w5 Munin", "Odin had two ravens Munin and Hugin", True),
		("Hugin |c5 Munin", "Odin had two ravens Munin and Hugin", True),
		("Hugin |w15 Munin", "", False),
		("Hugin |c15 Munin", "", False),
		("Munin |w0 Hugin", "Munin Hugin", True),
		("Munin |w0 Hugin", "Hugin Munin", True),
		("Munin |w0 Hugin", "Munin and Hugin", False),
		("T |c0 wo", "Two", True),
		("Hello world |w0 how are you", "Hello world how are you", True),
		("how are you |w0 Hello world", "Hello world how are you", True),
	]
)
def test_distance_function_results(condition, text, expected_result):
	assert Distance(condition, text).res == expected_result



from .execute import execute_condition

import pytest


@pytest.mark.parametrize(
	"condition, text, expected_result",
	[
		("odin |w5 great warrior", "many peoples think that odin was a great warrior", True),
	]
)
def test_execute_condition(condition, text, expected_result):
	assert execute_condition(condition, text) == expected_result



from .functions.auto_execute import execute_automatically
from .parse import parse_items
from .checkers.all_ import check_all

from typing import List, Union


def execute_condition(condition: str, text: str) -> bool:
	check_all(condition)
	condition = condition.replace(" И ", " AND ").replace(" ИЛИ ", " OR ")
	condition_parts = parse_items(condition)
	res = execute_condition_parts(condition_parts, text)

	return res


def execute_condition_parts(condition_parts: List[List[Union[str, int]]], text: str) -> bool:
	text = text.lower()
	result = False

	for i in range(len(condition_parts) - 1, -1, -1):
		part = condition_parts[i]
		condition = part[0] if len(part) == 1 else "".join(part)

		_res = "TRUE" if execute_condition_part(condition, text) else "FALSE"		# Converts to atomic

		if i > 0:		# I'm not sure about this condition
			condition_part_id = find_place_to_insert_result(condition_parts, i)
			atom_id = condition_parts[condition_part_id].index(i)

			condition_parts[condition_part_id][atom_id] = _res

		else:
			result = True if _res == "TRUE" else False

	return result


def execute_condition_part(condition: str, text: str):
	res = True

	if " AND " in condition:
		condition_parts = condition.split(" AND ")

		for part in condition_parts:
			res &= execute_condition_part(part, text)

		return res

	if "NOT " in condition:
		condition_part = condition.split("NOT ")[1]
		return not execute_condition_part(condition_part, text)

	if " OR " in condition:
		condition_parts = condition.split(" OR ")
		res = False

		for part in condition_parts:
			res |= execute_condition_part(part, text)

		return res

	return execute_automatically(condition, text)


def find_place_to_insert_result(condition_parts: List[List[str]], command_id: int):
	for i in range(len(condition_parts)):
		if command_id in condition_parts[i]:
			return i

	raise IndexError("Can't find command with this id")



from ..exceptions import ForsetiSyntaxError
from .syntax import check_syntax

import pytest


@pytest.mark.parametrize(
	"condition",
	[
		"AND SomeCommand",
		"SomeCommand AND",
		"OR SomeCommand",
		"SomeCommand OR",
		"И SomeCommand",
		"SomeCommand И",
		"ИЛИ SomeCommand",
		"SomeCommand ИЛИ",
		"NOT SomeCommand",
		"SomeCommand NOT",
	]
)
def test_check_syntax_errors(condition):
	with pytest.raises(ForsetiSyntaxError):
		check_syntax(condition)


@pytest.mark.parametrize(
	"condition",
	[
		"SomeCommand OR SomeCondition (Wrong condition)",
		"(Wrong condition) SomeCommand OR SomeCondition",
		"SomeCommand (Wrong condition) SomeCondition",
	]
)
def test_check_syntax_errors(condition):
	with pytest.raises(ForsetiSyntaxError, match="You can't use staples without operators"):
		check_syntax(condition)


@pytest.mark.parametrize(
	"condition",
	[
		"SomeCommand1 OR NOT SomeCommand2",
		"SomeCommand1 NOT NOT SomeCommand2",
		"SomeCommand1 NOT SomeCommand2",
	]
)
def test_check_staples_errors(condition):
	with pytest.raises(ForsetiSyntaxError, match="Operator 'NOT' cannot be used without operator 'AND'"):
		check_syntax(condition)



from .distance import Distance
from .length import Length
from .nearby import Nearby
from .regex import Regex

from re import search


def execute_automatically(command: str, text: str) -> bool:
	if command == "TRUE":		# ATOM
		return True

	elif command == "FALSE": 	# ATOM
		return False

	elif "|l" in command:
		return Length(command, text).res

	elif search(Distance.COMMAND_FORMAT, command):
		return Distance(command, text).res

	elif "|nearby " in command:
		return Nearby(command, text).res

	else:
		return Regex(command, text).res



from typing import List
from .exceptions import ForsetiSyntaxError


def parse_items(condition: str) -> List[List[str]]:
	buffer = ""
	stack = [[]]
	conditions = list()
	conditions.append(stack[-1])

	for char in condition:
		if char == "(":
			if not stack:
				raise ForsetiSyntaxError("The closing bracket is met before using the opening one")

			if buffer:
				stack[-1].append(buffer[:])

			stack.append([])
			stack[-2].append(len(conditions))
			conditions.append(stack[-1])

			buffer = ""

		elif char == ")":
			if not stack:
				raise ForsetiSyntaxError("The closing bracket is met before using the opening one")

			if buffer:
				stack[-1].append(buffer[:])

			del stack[-1]

			buffer = ""

		else:
			buffer += char

	if buffer:
		conditions[0].append(buffer)

	return conditions



from .length import Length
from ..exceptions import ForsetiFunctionSyntaxError

import pytest


@pytest.mark.parametrize(
	"condition, text",
	[
		("|lm", ""),
		("|lg", ""),
		("|ll", ""),
		("ll", ""),
		("|lgN", ""),
		("", ""),
	]
)
def test_length_function_exceptions(condition, text):
	with pytest.raises(ForsetiFunctionSyntaxError):
		Length(condition, text)


@pytest.mark.parametrize(
	"condition, text, expected_result",
	[
		("|ll5", "Odin had two ravens Hugin and Munin", False),
		("|lg5", "Odin had two ravens Hugin and Munin", True),
		("|ll15", "Hugin or Munin", True),
		("|lg15", "Hugin or Munin", False),
	]
)
def test_length_function_results(condition, text, expected_result):
	assert Length(condition, text).res == expected_result



from .base import BaseFunction

from re import search


class Regex(BaseFunction):
	def __init__(self, command: str, text: str):
		super().__init__(command, text)

	def execute(self) -> bool:
		return bool(search(self.command, self.text))



from .execute import execute_condition


def start():
    print("Welcome to Forseti! To exit, keep silent twice.")

    while True:
        condition = input("Condition: ")
        text = input("Text: ")

        if not any((condition, text)):
            print("Thanks for using!")
            exit(0)

        try:
            res = execute_condition(condition, text)
            print("Result:", res)

        except Exception as err:
            print("Error:", err)


if __name__ == "__main__":
    start()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['forseti_lang', 'forseti_lang.checkers', 'forseti_lang.functions']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['main = forseti.__main__:start']}

setup_kwargs = {
    'name': 'forseti-lang',
    'version': '1.0.1',
    'description': 'The system for determining the compliance of the text with the specified rules',
    'long_description': '# Forseti Language\n\nThe system for determining the compliance of the text with the specified rules\n\n## How to install:\n\nPyPi installation guide\n```shell\npip install forseti-lang\n```\n\n## How to use:\n\n```python\nfrom forseti_lang.execute import execute_condition\n\ncondition = "TRUE AND FALSE"\ntext = ""\nexecute_condition(condition, text)  # False\n```\n\nAll regular expressions should be in lowercase.\n\n## Supported logic operators\n\n### English\n\nBelow is a table of available logical operators and their description.\n\n| Operator               | Description                                 | Priority |\n|------------------------|---------------------------------------------|----------|\n| ()                     | The conditions in quotes will be met first  | 0        |\n| AND                    | Logical **AND**                             | 1        |\n| AND NOT                | Logical **AND NOT**                         | 2        |\n| OR                     | Logical **OR**                              | 3        |\n| Atoms (TRUE and FALSE) | Minimum logical unit                        | 4        |\n\n### Russian\n\nForseti was developed for the analysis of texts in Russian.\nSo, you can use:\n\n| RU  | ENG |\n|-----|-----|\n| И   | AND |\n| ИЛИ | OR  |\n\n#### Note\n\nYou can combine ru and eng operators.\n\n## Supported functions\n\nWe really lacked a simple syntax (but regular expressions did not allow us to achieve this), so we created our own\nfunctions!\n\n| Function         | In Forseti | Description                                               | Priority |\n|------------------|------------|-----------------------------------------------------------|----------|\n| f text length    | ll, lg     | Text length in characters check function                  | 5        |\n| f words distance | wN, cN, N  | Checks whether the words are in the specified interval    | 6        |\n| f words nearby   | nearby     | Checks the words that are next to another word            | 7        |\n| f regex search   |            | Checks the occurrence of a regular expression in the text | 8        |\n\n### Explanations\n\n* `|ll` - length less (in characters)\n* `|lg` - length great (in characters)\n* `|wN` - word in distance `N words`\n* `|cN` - word in distance `N characters`\n* `|N` - alias of `|wN`\n* `|nearby` - the word is nearby to the words `word1 |nearby word2 | word3 | word4`\n\n#### Notes\n\n* All function starts with special character: `|`. That\'s why we don\'t support it in regular expressions\n* It is also forbidden to use the `-` character, it is used in the distance function\n* The function of finding words at a distance will allow you to exclude constructions with a word at a distance,\nfor example:\n```\ncondition - "hello |w1 -world"\ntext      - "Hello beautiful world!"\nresult    - False\n```\n* You cannot combine two functions in one expression\n* Text length and word distance functions don\'t work with negative values\n\n## Priority table\n\n| Operator                   | Description                                               | Priority |\n|----------------------------|-----------------------------------------------------------|----------|\n| `()`                       | The conditions in quotes will be met first                | 0        |\n| `AND`                      | Logical `AND`                                             | 1        |\n| `AND NOT`                  | Logical `AND NOT`                                         | 2        |\n| `OR`                       | Logical `OR`                                              | 3        |\n| Atoms (`TRUE` and `FALSE`) | Minimum logical unit                                      | 4        |\n| f text length              | Text length in characters check function                  | 5        |\n| f words distance           | Checks whether the words are in the specified interval    | 6        |\n| f words nearby             | Checks the words that are next to another word            | 7        |\n| f regex search             | Checks the occurrence of a regular expression in the text | 8        |\n\n\n## Examples\n\n### Checking the length of the text\n\n```python\nfrom forseti_lang.execute import execute_condition\n\ncondition = \'|ll5\'\nexecute_condition(condition, "Hello world!")  # False\nexecute_condition(condition, "Hi")  # True\n\ncondition = \'|lg5\'\nexecute_condition(condition, "Hello world!")  # True\nexecute_condition(condition, "Hi")  # False\n```\n\n### Checking the words distance\n\n```python\nfrom forseti_lang.execute import execute_condition\n\ncondition = \'hello |w1 world\'\nexecute_condition(condition, "Hello world!")  # True\nexecute_condition(condition, "Hello sunny world!")  # True\nexecute_condition(condition, "Hello dirty and dark world!")  # False\n\ncondition = \'hi\\W? |c1 how are you\'\nexecute_condition(condition, "Hi, how are you?")  # True\nexecute_condition(condition, "Hi, hmm.. how ary you?")  # False\n\ncondition = \'hello |1 world\'\nexecute_condition(condition, "Hello world!")  # True\n```\n\n### Checking nearby words\n\n```python\nfrom forseti_lang.execute import execute_condition\n\ncondition = \'hello |nearby world | people | notalib\'\nexecute_condition(condition, "Hello world!")  # True\nexecute_condition(condition, "Hello notalib!")  # True\n```\n\n### Logic processing\n\n```python\nfrom forseti_lang.execute import execute_condition\n\nexecute_condition("TRUE AND FALSE", "") # False\nexecute_condition("TRUE OR FALSE", "") # True\nexecute_condition("FALSE OR TRUE AND (TRUE OR FALSE)", "") # True\nexecute_condition("FALSE OR TRUE AND (FALSE AND TRUE)", "") # False\nexecute_condition("(TRUE OR FALSE) AND NOT (TRUE AND FALSE)", "") # TRUE\n```\n\n### Difficult rules example\n* `короб AND NOT короб |2 конфет`\n* `минимальн\\w,{,2} |1 стоимо`\n* `балл AND NOT (сн[ия]\\w* балл ИЛИ 10\\s?балл ИЛИ десять\\sбаллов) AND |lg15`\n',
    'author': 'Kirill_Lekhov',
    'author_email': 'Kirill.Lekhov@mail.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Kirill-Lekhov/forseti_lang',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



from .base import BaseFunction
from .distance import Distance
from ..exceptions import ForsetiFunctionSyntaxError

from re import search, split as re_split


class Nearby(BaseFunction):
	def __init__(self, command: str, text: str):
		self.arguments = []
		self.left_side_argument = ""
		super().__init__(command, text)

	def check_command(self):
		if "|nearby" not in self.command:
			raise ForsetiFunctionSyntaxError("Unsupported command syntax")

		command_parts = self.command.split(" |nearby ")

		if len(command_parts) != 2:
			raise ForsetiFunctionSyntaxError("Command doesn't support combined-condition")

		self.left_side_argument, self.arguments = command_parts
		self.arguments = self.arguments.split(" | ")

	def execute(self) -> bool:
		if not search(self.left_side_argument, self.text):
			return False

		results = []

		for i in self.arguments:
			condition = self.left_side_argument + " |w0 " + i
			results.append(Distance(condition, self.text).res)

		return any(results)

	def _word_in_left_side(self, left_side_part: str, word: str) -> bool:
		pass

	def _word_in_right_side(self, right_side_part: str, word: str) -> bool:
		pass



from ..exceptions import ForsetiSyntaxError
from .staples import check_staples

import pytest


@pytest.mark.parametrize(
	"condition",
	[
		"((True)",
		"(True))",
	]
)
def test_check_staples_errors(condition):
	with pytest.raises(ForsetiSyntaxError):
		check_staples(condition)


@pytest.mark.parametrize(
	"condition",
	[
		"(True)",
		"((True))",
		"True",
		"(True OR False) AND (False AND True)",
	]
)
def test_check_staples_results(condition):
	assert check_staples(condition) is None



from .base import BaseFunction
from ..exceptions import ForsetiFunctionSyntaxError

from typing import Tuple
from re import search, split as re_split, finditer


class Distance(BaseFunction):
	COMMAND_FORMAT = r'\s\|[cw]?\d+\s'

	def __init__(self, command: str, text: str):
		self.left_argument = ""
		self.right_argument = ""
		self.distance = -1
		self.f_exclude_right_argument = False

		super().__init__(command, text)

	def check_command(self):
		if bool(search(self.COMMAND_FORMAT, self.command)):
			parts = re_split(self.COMMAND_FORMAT, self.command)

			if len(parts) != 2:
				raise ForsetiFunctionSyntaxError("Command doesn't support combined-condition")

			command_to_exec = search(self.COMMAND_FORMAT, self.command).group(0).strip()

			if 'c' in command_to_exec or 'w' in command_to_exec:
				_p = 2
			else:
				_p = 1

			self.distance = int(command_to_exec[_p:])
			self.left_argument, self.right_argument = parts

			if self.right_argument[0] == '-':
				self.right_argument = self.right_argument[1:]
				self.f_exclude_right_argument = True

		else:
			raise ForsetiFunctionSyntaxError("Unsupported command format. You must use: {word1} [|c{number} or |w{number}] {word2}")

	def execute(self) -> bool:
		if "|w" in self.command:
			return self._by_words()

		elif "|c" in self.command:
			return self._by_characters()

		else:
			return self._by_words()

	def _by_words(self) -> bool:
		if not self._parts_in_text():
			return False

		text_by_words = self.text.split()
		words_number = len(text_by_words)
		words_in_arguments = len(self.left_argument.split()) + len(self.right_argument.split())

		for i in range(words_number):
			text_part = ' '.join(text_by_words[i:i+words_in_arguments+self.distance])

			if search(self.left_argument, text_part) and search(self.right_argument, text_part):
				return True

		return False

	def _by_characters(self) -> bool:
		if not self._parts_in_text():
			return False

		def check_pair_distance(pair_1: Tuple[int, int], pair_2: Tuple[int, int]) -> bool:
			pairs = ((pair_1[0], pair_2[1]), (pair_1[1], pair_2[0]))

			if max(pairs[0]) - min(pairs[0]) <= self.distance or max(pairs[1]) - min(pairs[1]) <= self.distance:
				return True

			return False

		left_argument_positions = list(map(lambda x: (x.start(), x.end()), finditer(self.left_argument, self.text)))
		right_argument_positions = list(map(lambda x: (x.start(), x.end()), finditer(self.right_argument, self.text)))

		for i in left_argument_positions:
			if any(map(lambda x: check_pair_distance(i, x), right_argument_positions)):
				return True

		return False

	def _parts_in_text(self) -> bool:
		return bool(search(self.left_argument, self.text) and search(self.right_argument, self.text))



from .parse import parse_items
from .exceptions import ForsetiSyntaxError

import pytest


@pytest.mark.parametrize(
	"condition",
	[
		"TRUE AND )FALSE(",
		"TRUE AND (()FALSE))(",
	]
)
def test_parse_items_exceptions(condition):
	with pytest.raises(ForsetiSyntaxError, match="The closing bracket is met before using the opening one"):
		parse_items(condition)


@pytest.mark.parametrize(
	"condition, expected_result",
	[
		("TRUE AND ((FALSE))", [['TRUE AND ', 1], [2], ['FALSE']]),
		("TRUE AND FALSE OR (TRUE AND (FALSE OR TRUE))", [['TRUE AND FALSE OR ', 1], ['TRUE AND ', 2], ['FALSE OR TRUE']]),
		("TRUE AND FALSE", [['TRUE AND FALSE']]),
		("", [[]])
	]
)
def test_parse_items_results(condition, expected_result):
	assert parse_items(condition) == expected_result



from .nearby import Nearby
from ..exceptions import ForsetiFunctionSyntaxError

import pytest


@pytest.mark.parametrize(
	"condition, text",
	[
		("|nearby", ""),
		(" |nearby ", ""),
		("", ""),
	]
)
def test_nearby_function_exceptions(condition, text):
	with pytest.raises(ForsetiFunctionSyntaxError):
		Nearby(condition, text)


@pytest.mark.parametrize(
	"condition, text, expected_result",
	[
		("Hugin, |nearby Munin", "Odin had two ravens Hugin, Munin", True),
		("Odin had son. And he was |nearby Tor | Vali | Vidar | Heimdall | Bragi | Tur", "Odin had son. And he was Tor", True),
		("Odin had son. And he was |nearby Tor | Vali | Vidar | Heimdall | Bragi | Tur", "Odin had son. And he was UNKNOWN", False),
		("elephant |nearby buy | well buy", "buy elephant", True),
		("elephant |nearby buy | well buy", "elephant buy", True),
		("elephant |nearby buy | well buy", "well buy elephant", True),
		("elephant |nearby buy | well buy", "elephant well buy", True),
	]
)
def test_nearby_function_results(condition, text, expected_result):
	assert Nearby(condition, text).res == expected_result
